# Apache License Version 2.0
# Copyright 2022 Xin Huang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse, os, sys, signal


def _set_sigpipe_handler():
    if os.name == "posix":
        # Set signal handler for SIGPIPE to quietly kill the program.
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)


def _run_score(args):
    from sstar.cal_s_star import cal_s_star
    cal_s_star(vcf=args.vcf, ref_ind_file=args.ref_ind, tgt_ind_file=args.tgt_ind, anc_allele_file=args.anc_allele, win_len=args.win_len, win_step=args.win_step, output=args.output, thread=args.thread, match_bonus=args.match_bonus, max_mismatch=args.max_mismatch, mismatch_penalty=args.mismatch_penalty)


def _run_quantile(args):
    from sstar.get_quantile import get_quantile
    get_quantile(model=args.model, ms_dir=args.ms_dir, N0=args.N0, nsamp=args.nsamp, nreps=args.nreps, ref_index=args.ref_index, ref_size=args.ref_size, tgt_index=args.tgt_index, tgt_size=args.tgt_size, mut_rate=args.mut_rate, rec_rate=args.rec_rate, seq_len=args.seq_len, snp_num_range=args.snp_num_range, output_dir=args.output_dir, thread=args.thread, seeds=args.seeds)


def _run_threshold(args):
    from sstar.cal_threshold import cal_threshold
    cal_threshold(simulated_data=args.sim_data, score_file=args.score, recomb_rate=args.recomb_rate, recomb_map=args.recomb_map, quantile=args.quantile, output=args.output, k=args.k)


def _run_match_pct(args):
    from sstar.cal_match_rate import cal_match_pct
    cal_match_pct(vcf=args.vcf, ref_ind_file=args.ref_ind, tgt_ind_file=args.tgt_ind, src_ind_file=args.src_ind, anc_allele_file=args.anc_allele, output=args.output, thread=args.thread, score_file=args.score, mapped_region_file=args.mapped_region_file)


def _run_tract(args):
    from sstar.get_tract import get_tract
    get_tract(threshold_file=args.threshold, match_pct_files=args.match_pct, output_prefix=args.output, diff=args.diff)


def _add_common_args(parser):
    parser.add_argument('--anc-allele', type=str, dest='anc_allele', default=None, help='name of the BED format file containing ancestral allele information, otherwise assuming the REF allele is the ancestral allele and the ALT allele is the derived allele; default: None')
    parser.add_argument('--output', type=str, required=True, help='name of the output file')
    parser.add_argument('--thread', type=int, default=1, help='number of threads for multiprocessing; default: 1')


def _add_mapped_args(parser):
    parser.add_argument('--mapped-region', type=str, dest='mapped_region_file', help='name of the BED file containing mapped regions')


def _add_window_args(parser):
    parser.add_argument('--win-len', type=int, dest='win_len', default=50000, help='length of the window to calculate S* scores; default: 50000')
    parser.add_argument('--win-step', type=int, dest='win_step', default=10000, help='step size for moving windows along genomes; default: 10000')


def _add_ref_ind_args(parser):
    parser.add_argument('--ref', type=str, dest='ref_ind', required=True, help='name of the file containing information for samples without introgression')


def _add_tgt_ind_args(parser):
    parser.add_argument('--tgt', type=str, dest='tgt_ind', required=True, help='name of the file containing information for samples for detecting introgression')


def _add_src_ind_args(parser):
    parser.add_argument('--src', type=str, dest='src_ind', required=True, help='name of the file containing information for samples from source populations')


def _add_score_args(parser):
    parser.add_argument('--score', type=str, dest='score', required=True, default=None, help='name of the file containing S* scores calculated by `sstar score`')


def required_length(nmin,nmax):
    class RequiredLength(argparse.Action):
        def __call__(self, parser, args, values, option_string=None):
            if not nmin<=len(values)<=nmax:
                msg='argument "{f}" requires between {nmin} and {nmax} arguments'.format(
                    f=self.dest,nmin=nmin,nmax=nmax)
                raise argparse.ArgumentTypeError(msg)
            setattr(args, self.dest, values)
    return RequiredLength


def _s_star_cli_parser():
    top_parser = argparse.ArgumentParser()
    subparsers = top_parser.add_subparsers(dest="subcommand")
    subparsers.required = True

    # Arguments for score subcommand
    parser = subparsers.add_parser('score', help='calculate S* scores from VCF files')
    parser.add_argument('--vcf', type=str, dest='vcf', required=True, help='name of the VCF file containing genotypes from samples')
    _add_ref_ind_args(parser)
    _add_tgt_ind_args(parser)
    _add_common_args(parser)
    _add_window_args(parser)
    parser.add_argument('--match-bonus', type=int, dest='match_bonus', default=5000, help='bonus for matching genotypes of two different variants; default: 5000')
    parser.add_argument('--max-mismatch', type=int, dest='max_mismatch', default=5, help='maximum genotype distance allowed; default: 5')
    parser.add_argument('--mismatch-penalty', type=int, dest='mismatch_penalty', default=-10000, help='penalty for mismatching genotypes of two different variants; default: -10000')
    parser.set_defaults(runner=_run_score)

    # Arguments for quantile subcommand
    parser = subparsers.add_parser('quantile', help='calculate quantiles for S* scores from simulated data without introgression')
    parser.add_argument('--model', type=str, required=True, help='demographic model without introgression for simulation in Demes YAML format')
    parser.add_argument('--ms-dir', type=str, dest='ms_dir', required=True, help='directory for the ms program for simulation')
    parser.add_argument('--N0', type=int, required=True, help='N0 used in ms simulation')
    parser.add_argument('--nsamp', type=int, required=True, help='sample size (haploid) used in ms simulation')
    parser.add_argument('--nreps', type=int, required=True, help='number of replicates used in ms simulation')
    parser.add_argument('--seeds', type=int, nargs=3, default=None, help='three random seed numbers used in ms simulation; default: None')
    parser.add_argument('--ref-index', type=int, dest='ref_index', required=True, help='index of the reference population in the demographic model (start from 1)')
    parser.add_argument('--ref-size', type=int, dest='ref_size', required=True, help='sample size (haploid) of the reference population')
    parser.add_argument('--tgt-index', type=int, dest='tgt_index', required=True, help='index of the target population in the demographic model (start from 1)')
    parser.add_argument('--tgt-size', type=int, dest='tgt_size', required=True, help='sample size (haploid) of the target population')
    parser.add_argument('--mut-rate', type=float, dest='mut_rate', required=True, help='mutation rate per generation per base')
    parser.add_argument('--rec-rate', type=float, dest='rec_rate', required=True, help='recombination rate per generation per base')
    parser.add_argument('--seq-len', type=int, dest='seq_len', required=True, help='length of simulated sequence')
    parser.add_argument('--snp-num-range', type=int, dest='snp_num_range', nargs=3, required=True, help='range of SNP numbers in ms simulation; the first parameter is the minimum SNP number, the second parameter is the maximum SNP number, the third parameter is the step size')
    parser.add_argument('--output-dir', type=str, dest='output_dir', required=True, help='directory for the output files')
    parser.add_argument('--thread', type=int, default=1, help='number of thread')
    parser.set_defaults(runner=_run_quantile)
    
    # Arguments for threshold subcommand
    parser = subparsers.add_parser('threshold', help='calculate S* thresholds from simulated data')
    _add_score_args(parser)
    parser.add_argument('--sim-data', type=str, dest='sim_data', required=True, help='name of the file containing simulated data for building a generalized additive model')
    parser.add_argument('--recomb-rate', type=float, dest='recomb_rate', default=1e-8, help='a uniform recombination rate used across the genome; default: 1e-8')
    parser.add_argument('--recomb-map', type=str, dest='recomb_map', default=None, help='a recombination map used across the genome; default: None')
    parser.add_argument('--quantile', type=float, required=True, default=0.99, help='a quantile for determining significant S* scores; default: 0.99')
    parser.add_argument('--output', type=str, required=True, help='name of the output file')
    parser.add_argument('--k', type=int, default=8, help='dimension(s) of the bases used to represent the smooth term')
    parser.set_defaults(runner=_run_threshold)

    # Arguments for matchpct subcommand
    parser = subparsers.add_parser('matchrate', help='calculate source match rates in target populations with genomes from source populations')
    parser.add_argument('--vcf', type=str, dest='vcf', required=True, help='name of the VCF file containing genotypes from samples')
    _add_ref_ind_args(parser)
    _add_tgt_ind_args(parser)
    _add_src_ind_args(parser)
    _add_common_args(parser)
    _add_mapped_args(parser)
    _add_score_args(parser)
    parser.set_defaults(runner=_run_match_pct)

    # Arguments for tract subcommand
    parser = subparsers.add_parser('tract', help='get candidate introgressed fragments')
    parser.add_argument('--threshold', type=str, required=True, help='threshold file from `sstar threshold`')
    parser.add_argument('--match-rate', type=str, default=None, dest='match_pct', nargs='+', action=required_length(1,2), help='match rate files from `sstar matchrate`; the first file contains match percents using genomes from the source population 1 (src1), the second file contains match percents using genomes from the source population 2 (src2)')
    parser.add_argument('--output-prefix', type=str, dest='output', required=True, help='prefix for output files')
    parser.add_argument('--diff', type=float, default=0, help='difference between src1 match rates (src1_match_rate) and src2 match rates (src2_match_rate); if src1_match_rate - src2_match_rate > diff, then this fragment is assigned to src1, if src1_match_rate - src2_match_rate < diff, then this fragment is assigned to src2; default: 0')
    parser.set_defaults(runner=_run_tract)

    return top_parser


def main(arg_list=None):
    """
    Description:
        Main entry fo sstar

    Arguments:
        arg_list list: List containing arguments for sstar
    """
    _set_sigpipe_handler()
    parser = _s_star_cli_parser()
    args = parser.parse_args(arg_list)
    args.runner(args)



# Apache License Version 2.0
# Copyright 2022 Xin Huang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
import rpy2.robjects as ro
from rpy2.robjects import FloatVector, Formula, pandas2ri
from rpy2.robjects.packages import importr
from rpy2.robjects.conversion import localconverter

#@profile
def cal_threshold(simulated_data, score_file, recomb_rate, recomb_map, quantile, output, k):
    """
    Description:
        Calculate significant S* scores from simulated data.

    Arguments:
        simulated_data str: Name of the file containing simulated data.
        score_file str: Name of the file containing S* scores.
        recomb_rate float: Uniform recombination rate assumed across the genome.
        recomb_map str: Name of the file containing recombination maps. 
        quantile float: Quantile for estimating significant S* scores.
        output str: Name of the output file.
        k int: Dimension(s) of the bases used to represent the smooth term.
    """
    if (recomb_rate != None) and (recomb_map != None): fit_lr = True
    else: fit_lr = False

    gam = _build_gam_model(simulated_data, k, fit_lr)
    res = _predict_res(gam, score_file, recomb_rate, recomb_map, quantile, fit_lr)
    header = 'chrom\tstart\tend\tsample\tS*_score\texpected_S*_score\tlocal_recomb_rate\tquantile\tsignificant'
    with open(output, 'w') as o:
        o.write(header+"\n")
        o.write("\n".join(res)+"\n")

#@profile
def _build_gam_model(simulated_data, k, fit_lr):
    """
    Description:
        Helper function for building a generalized additive model with R package MGCV.

    Arguments:
        simulated_data str: Name of the file containing simulated data.
        k int: Dimension(s) of the bases used to represent the smooth term.
        fit_lr bool: If True, building a gam with local recombination rate.

    Returns:
        gam rpy2.robjects: Generalized additive model for prediction.
    """
    mgcv = importr('mgcv')

    data = pd.read_csv(simulated_data, sep="\t")

    s_star = FloatVector(data.iloc[:,0].values)
    snps = FloatVector(data.iloc[:,1].values)
    q = FloatVector(data.iloc[:,2].values)

    ro.globalenv['s_star'] = s_star
    ro.globalenv['snps'] = snps
    ro.globalenv['q'] = q

    if fit_lr: 
        lr = FloatVector(data.iloc[:,3].values)
        ro.globalenv['lr'] = lr
        fmla = f's_star ~ te(snps, lr, q, k={k})'
    else: fmla = f's_star ~ te(snps, q, k={k})'

    fmla = Formula(fmla)

    gam = mgcv.gam(fmla)

    return gam

#@profile
def _predict_res(gam, score_file, recomb_rate, recomb_map, quantile, fit_lr):
    """
    Description:
        Helper function to predict significant S* scores.

    Arguments:
        gam rpy2.robjects: Generalized additive model for prediction.
        score_file str: Name of the file containing S* scores.
        recomb_rate float: Uniform recombination rate assumed across the genome.
        recomb_map str: Name of the file containing recombination maps.
        quantile float: Quantile for estimating significant S* scores.
        fit_lr bool: If True, building a gam with local recombination rate.

    Returns:
        res list: List containing results for output.
    """
    stats = importr('stats')

    if recomb_map != None: recomb_map = _read_recomb_map(recomb_map)
    data, meta_data = _read_score_file(score_file, recomb_rate, recomb_map, quantile, fit_lr)

    thresholds = dict()
    for s in data.keys():
        if fit_lr: pd_df = pd.DataFrame({'snps': data[s]['snps'], 'lr': data[s]['lr'], 'q': data[s]['q']})
        else: pd_df = pd.DataFrame({'snps': data[s]['snps'], 'q': data[s]['q']})

        with localconverter(ro.default_converter + pandas2ri.converter):
            r_from_pd_df = ro.conversion.py2rpy(pd_df)
        
        thresholds[s] = np.array(stats.predict(gam, r_from_pd_df))

    res = []
    for s in meta_data.keys():
        for i in range(len(meta_data[s])):
            score = float(meta_data[s][i][4])
            null_score = round(thresholds[s][i], 6)
            if score > null_score: sig = 'True'
            else: sig = 'False'

            line = "\t".join(meta_data[s][i])
            line += "\t" + str(null_score)
            line += "\t" + str(10**data[s]['lr'][i])
            line += "\t" + str(data[s]['q'][i])
            line += "\t" + sig

            res.append(line)

    return res

#@profile
def _read_score_file(score_file, recomb_rate, recomb_map, quantile, fit_lr):
    """
    Description:
        Helper function for reading files from `sstar score`.

    Arguments:
        score_file str: Name of the file containing S* scores.
        recomb_rate float: Uniform recombination rate assumed across the genome. 
        recomb_map str: Name of the file containing recombination maps.
        quantile float: Quantile for estimating significant S* scores.
        fit_lr bool: If True, building a gam with local recombination rate.

    Returns:
        data dict: Dictionary containing data for predicting significat S* scores.
        meta_data dict: Dictionary containing meta data for output.
    """
    data = dict()
    meta_data = dict()
    with open(score_file, 'r') as f:
        f.readline()
        for line in f.readlines():
            line = line.rstrip()
            element = line.split("\t")
            if element[6] == 'NA': continue
            else:
                chr_name = element[0]
                win_start = element[1]
                win_end = element[2]
                sample = element[3]
                score = element[4]
                total_snp_num = element[5]
                if fit_lr:
                    if recomb_map != None: 
                        key = chr_name+":"+win_start+"-"+win_end
                        if key in recomb_map.keys(): local_recomb_rate = np.log10(recomb_map[key])
                        else: continue
                    else:
                        local_recomb_rate = np.log10(recomb_rate)
                else: local_recomb_rate = np.nan

                if sample not in data.keys(): 
                    data[sample] = dict()
                    data[sample]['snps'] = list()
                    data[sample]['lr'] = list()
                    data[sample]['q'] = list()
                if sample not in meta_data.keys(): meta_data[sample] = list()
                data[sample]['snps'].append(float(total_snp_num))
                data[sample]['lr'].append(local_recomb_rate)
                data[sample]['q'].append(quantile)
                meta_data[sample].append([chr_name, win_start, win_end, sample, score])

    return data, meta_data

#@profile
def _read_recomb_map(recomb_map_file):
    """
    Description:
        Helper function for reading recombination maps from files.

    Arguments:
        recomb_map_file str: Name of the file containing recombination maps.

    Returns:
        recomb_map dict: Dictionary containing recombination maps.
    """
    recomb_map = dict()
    with open(recomb_map_file, 'r') as f:
        for line in f.readlines():
            line = line.rstrip()
            element = line.split("\t")
            key = element[0]+":"+element[1]+"-"+element[2]
            if key not in recomb_map.keys(): recomb_map[key] = float(element[3])

    return recomb_map

if __name__ == '__main__':
   cal_threshold(simulated_data="../examples/data/simulated_data/gravel_asn_scale_60k.simulated.data", score_file="../tests/data/test.score", recomb_rate=1.29, recomb_map=None, quantile=0.99, output="threshold.out")



# Apache License Version 2.0
# Copyright 2022 Xin Huang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import demes
import glob
import subprocess
import numpy as np
import pandas as pd
from multiprocessing import Process, Queue
from scipy.stats import norm
from scipy.stats import nbinom


def get_quantile(model, ms_dir, N0, nsamp, nreps, ref_index, ref_size, tgt_index, tgt_size, mut_rate, rec_rate, seq_len, snp_num_range, output_dir, thread, seeds):
    """
    Description:
        Calculates quantiles of expected S*.

    Arguments:
        model str: Name of file containing the demographic model for simulation.
        ms_dir str: Name of the directory containing the ms program.
        N0 int: N0 used in ms simulation.
        nsamp int: Sample size (haploid) used in ms simulation.
        nreps int: Number of replicates used in ms simulation.
        ref_index int: Index of the reference population in the demographic model (start from 1).
        ref_size int: Sample size (haploid) of the reference population.
        tgt_index int: Index of the target population in the demographic model (start from 1).
        tgt_size int: Sample size (haploid) of the target population.
        mut_rate float: Mutation rate.
        rec_rate float: Recombination rate.
        seq_len int: Length of simulated sequence.
        snp_num_range list: Range of SNP numbers in ms simulation; the first parameter is the minimum SNP number, the second parameter is the maximum SNP number, the third parameter is the step size.
        output_dir str: Number of the output directory.
        thread int: Number of threads.
        seeds: list: Three random seed numbers used in ms simulation.
    """
    if seeds is not None: np.random.seed(np.sum(seeds))
    output_dir = os.path.abspath(output_dir)
    if os.path.exists(output_dir) is False: subprocess.call(['mkdir', output_dir])
    _generate_mut_rec_combination(N0, nreps, mut_rate, rec_rate, seq_len, output_dir)
    _run_ms_simulation(model, ms_dir, N0, nsamp, nreps, ref_index, ref_size, tgt_index, tgt_size, seq_len, snp_num_range, output_dir, thread, seeds)
    _summary(output_dir, rec_rate)


def _generate_mut_rec_combination(N0, nreps, mut_rate, rec_rate, seq_len, output_dir):
    """
    Description:
        Helper function to create different combination of mutation rates and recombination rates.

    Arguments:
        N0 int: N0 used in ms simulation.
        nreps int: Number of replicates used in ms simulation.
        mut_rate float: Mutation rate.
        rec_rate float: Recombination rate.
        seq_len int: Length of simulated sequence.
        output_dir str: Name of the output directory.
    """
    scaled_mut_rate = 4*N0*mut_rate*seq_len
    scaled_rec_rate = 4*N0*rec_rate*seq_len
    mut_rate_list = norm.rvs(loc=scaled_mut_rate, scale=0.233, size=nreps)
    rec_rate_list = nbinom.rvs(n=0.5, p=0.5/(0.5+scaled_rec_rate), size=nreps)

    rates = f'{output_dir}/rates.combination'
    with open(rates, 'w') as o:
        for i in range(len(mut_rate_list)):
            if mut_rate_list[i] < 0.001: mut_rate_list[i] = 0.001
            if rec_rate_list[i] < 0.001: rec_rate_list[i] = 0.001
            mut_rate = mut_rate_list[i]
            rec_rate = rec_rate_list[i]
            o.write(f'{mut_rate}\t{rec_rate}\n')


def _run_ms_simulation(model, ms_dir, N0, nsamp, nreps, ref_index, ref_size, tgt_index, tgt_size, seq_len, snp_num_range, output_dir, thread, seeds):
    """
    Description
        Helper function for running ms simulation.

    Arguments:
        model str: Name of file containing the demographic model for simulation.
        ms_dir str: Name of the directory containing the ms program.
        N0 int: N0 used in ms simulation.
        nsamp int: Sample size (haploid) used in ms simulation.
        nreps int: Number of replicates used in ms simulation.
        ref_index int: Index of the reference population in the demographic model (start from 1).
        ref_size int: Sample size (haploid) of the reference population.
        tgt_index int: Index of the target population in the demographic model (start from 1).
        tgt_size int: Sample size (haploid) of the target population.
        seq_len int: Length of simulated sequence.
        snp_num_range list: Range of SNP numbers in ms simulation; the first parameter is the minimum SNP number, the second parameter is the maximum SNP number, the third parameter is the step size.
        output_dir str: Name of the output directory.
        thread int: Number of threads.
        seeds list: Three random seed numbers used in ms simulation.
    """
    graph = demes.load(model)
    samples = np.zeros(len(graph.demes))
    samples[ref_index-1] = ref_size
    samples[tgt_index-1] = tgt_size
    ms_params = demes.to_ms(graph, N0=N0, samples=samples)
    snp_num_list = np.arange(snp_num_range[0], snp_num_range[1]+snp_num_range[2], snp_num_range[2])
    
    ms_exec = os.path.abspath(ms_dir) + '/ms'
    rates = f'{output_dir}/rates.combination'
    ref_list = f'{output_dir}/sim.ref.list'
    tgt_list = f'{output_dir}/sim.tgt.list'
    ref_size = int(ref_size / 2)
    tgt_size = int(tgt_size / 2)

    if ref_index == tgt_index: raise Exception('The reference population should be different from the target population.')
    elif ref_index < tgt_index:
        with open(ref_list, 'w') as o:
            for i in range(ref_size):
                o.write(f'ms_{i}\n')
        with open(tgt_list, 'w') as o:
            for i in range(tgt_size):
                o.write(f'ms_{i+ref_size}\n')
    else:
        with open(ref_list, 'w') as o:
            for i in range(ref_size):
                o.write(f'ms_{i+tgt_size}\n')
        with open(tgt_list, 'w') as o:
            for i in range(tgt_size):
                o.write(f'ms_{i}\n')
       
    try:
        from pytest_cov.embed import cleanup_on_sigterm
    except ImportError:
        pass
    else:
        cleanup_on_sigterm()

    in_queue, out_queue = Queue(), Queue()
    workers = [Process(target=_run_ms_simulation_worker, args=(in_queue, out_queue, output_dir, rates, ms_exec, nsamp, nreps, seq_len, ms_params, ref_list, tgt_list, seeds)) for ii in range(thread)]
 
    for snp_num in snp_num_list:
        in_queue.put(snp_num)

    try:
        for worker in workers:
            worker.start()
        for snp_num in snp_num_list:
            out_queue.get()
        for worker in workers:
            worker.terminate()
    finally:
        for worker in workers:
            worker.join()


def _run_ms_simulation_worker(in_queue, out_queue, output_dir, rates, ms_exec, nsamp, nreps, seq_len, ms_params, ref_list, tgt_list, seeds):
    """
    Description:
        Worker function for running ms simulation.

    Arguments:
        in_queue multiprocessing.Queue: multiprocessing.Queue instance to receive parameters from the manager.
        out_queue multiprocessing.Queue: multiprocessing.Queue instance to send results back to the manager.
        output_dir str: Name of the output directory.
        rates str: Name of the file containing different combination of mutation rates and recombination rates.
        ms_exec str: Path to the ms program.
        nsamp int: Sample size (haploid) used in ms simulation.
        nreps int: Number of replicates used in ms simulation.
        seq_len int: Length of the simulated sequeuence.
        ms_params list: List of ms parameters.
        ref_list str: Name of the file containing individuals from the reference population.
        tgt_list str: Name of the file containing individuals from the target population.
        seeds list: Three random seed numbers used in ms simulation.
    """
    while True:
        snp_num = in_queue.get()
        output_subdir = f'{output_dir}/{snp_num}'
        output_ms = f'{output_subdir}/sim.ms'
        output_vcf = f'{output_subdir}/sim.vcf'
        output_score = f'{output_subdir}/sim.score'
        output_quantile = f'{output_subdir}/sim.quantile'
        ms_script = f'{output_subdir}/run_ms.sh'

        if seeds is not None:
            cmd = " ".join(['cat', rates, '|', ms_exec, str(nsamp), str(nreps), '-seeds', " ".join([str(s) for s in seeds]), '-t', 'tbs', '-r', 'tbs', str(seq_len), '-s', str(snp_num), ms_params, '>', output_ms])
        else:
            cmd = " ".join(['cat', rates, '|', ms_exec, str(nsamp), str(nreps), '-t', 'tbs', '-r', 'tbs', str(seq_len), '-s', str(snp_num), ms_params, '>', output_ms])
        
        if os.path.exists(output_subdir) is False: subprocess.call(['mkdir', output_subdir])
        with open(ms_script, 'w') as o:
            o.write(cmd+"\n")
        subprocess.call(['bash', ms_script])
        _ms2vcf(output_ms, output_vcf, nsamp, seq_len)
        subprocess.call(['sstar', 'score', '--vcf', output_vcf, '--ref', ref_list, '--tgt', tgt_list, '--output', output_score, '--win-len', str(seq_len), '--win-step', str(seq_len), '--thread', '1'])
        _cal_quantile(output_score, output_quantile, snp_num)
        out_queue.put('Finished')


def _ms2vcf(ms_file, vcf_file, nsamp, seq_len, ploidy=2):
    """
    Description:
        Converts ms output files into the VCF format.

    Arguments:
        ms_file str: Name of the ms file (input).
        vcf_file str: Name of the VCF file (output).
        nsamp int: Number of haploid genomes.
        seq_len int: Sequence length.
        ploidy int: Ploidy of each individual.
    """
    data = []
    i = -1
    header = "##fileformat=VCFv4.2\n"
    header += "##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">\n"
    header += "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t" + "\t".join(['ms_' + str(i) for i in range(int(nsamp/ploidy))])

    with open(ms_file, 'r') as f:
        f.readline()
        f.readline()
        for l in f.readlines():
            if l.startswith('//'):
                i += 1
                data.append({})
                data[i]['pos'] = []
                data[i]['geno'] = []
            elif l.startswith('positions'):
                data[i]['pos'] = l.rstrip().split(" ")[1:]
            elif l.startswith('0') or l.startswith('1'):
                data[i]['geno'].append(l.rstrip())

    shift = 0
    with open(vcf_file, 'w') as o:
        o.write(header+"\n")
        for i in range(len(data)):
            for j in range(len(data[i]['pos'])):
                pos = int(seq_len * float(data[i]['pos'][j])) + shift
                genotypes = "".join([data[i]['geno'][k][j] for k in range(len(data[i]['geno']))])
                genotypes = "\t".join([a+'|'+b for a,b in zip(genotypes[0::ploidy],genotypes[1::ploidy])])
                o.write(f"1\t{pos}\t.\tA\tT\t100\tPASS\t.\tGT\t{genotypes}\n")
            shift += seq_len


def _cal_quantile(in_file, out_file, snp_num):
    """
    Description:
        Helper function for calculating quantiles of expected S* with a given SNP number.

    Arguments:
        in_file str: Name of the input file containing S* scores.
        out_file str: Name of the output file.
        snp_num int: Number of SNPs used in the simulation.
    """
    df = pd.read_csv(in_file, sep="\t").dropna()
    quantiles = np.arange(0.5,1,0.005)
    mean_df = df.groupby(['chrom', 'start', 'end'], as_index=False)['S*_score'].mean().dropna()
    scores = np.quantile(mean_df['S*_score'], quantiles)
    with open(out_file, 'w') as o:
        o.write('S*_score\tSNP_num\tquantile\n')
        for i in range(len(scores)):
            o.write(f'{scores[i]}\t{snp_num}\t{quantiles[i]}\n')


def _summary(output_dir, rec_rate):
    """
    Description:
        Helper function for summarize quantiles of expected S* from different SNP numbers.

    Arguments:
        output_dir str: Name of the output directory.
        rec_rate float: Recombination rate.
    """
    all_files = glob.glob(f'{output_dir}/*/*.quantile')
    li = []
    for filename in all_files:
        df = pd.read_csv(filename, sep='\t')
        li.append(df)

    df = pd.concat(li, ignore_index=True).round(3)
    df['log(local_recomb_rate)'] = np.log10(rec_rate)
    df.sort_values(by=['SNP_num', 'quantile']).to_csv(f'{output_dir}/quantile.summary.txt', sep='\t', index=False)



import os.path
from setuptools import setup

# The directory containing this file
HERE = os.path.abspath(os.path.dirname(__file__))

# The text of the README file
with open(os.path.join(HERE, "README.md")) as fid:
    README = fid.read()

# This call to setup() does all the work
setup(
    name="sstar",
    version="1.1.1",
    description="A Python package for detecting archaic introgressin from population genetic data with S*",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/xin-huang/sstar",
    author="Xin Huang",
    author_email="xinhuang.res@gmail.com",
    license="Apache-2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    packages=["sstar"],
    include_package_data=True,
    install_requires=[
        "pandas",
        "scikit-allel",
        "rpy2"
    ],
    entry_points={"console_scripts": ["sstar=sstar.__main__:main"]},
)



# Apache License Version 2.0
# Copyright 2022 Xin Huang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd


def get_tract(threshold_file, match_pct_files, output_prefix, diff):
    """
    Description:
        Gets candidate introgressed fragments.

    Arguments:
        threshold_file str: Name of the file containing thresholds.
        match_pct_files list: List containing names of files containing source match rates.
        output_prefix str: Prefix of the output files.
        diff float: Cut-off to determine the origin of the candidate introgressed fragments.
    """
    threshold_df = pd.read_csv(threshold_file, sep="\t")
    threshold_df = threshold_df[threshold_df['significant'] == True]

    if match_pct_files is None:
        _output_bed(threshold_df, output_prefix+'.bed')
    elif len(match_pct_files) == 1:
        src_df = pd.read_csv(match_pct_files[0], sep="\t")
        merge_df = pd.merge(src_df, threshold_df, on=['chrom', 'start', 'end', 'sample'])
        _output_bed(merge_df, output_prefix+'.bed', col='match_rate')
    else:
        src1_df = pd.read_csv(match_pct_files[0], sep="\t")
        src2_df = pd.read_csv(match_pct_files[1], sep="\t")
        src1_sig_df = pd.merge(src1_df, threshold_df, on=['chrom', 'start', 'end', 'sample'])
        src2_sig_df = pd.merge(src2_df, threshold_df, on=['chrom', 'start', 'end', 'sample'])
        df = pd.merge(src1_sig_df, src2_sig_df, on=['chrom', 'start', 'end', 'sample'])
        src1_df = df[df['match_rate_x']-df['match_rate_y']>diff]
        src2_df = df[df['match_rate_x']-df['match_rate_y']<diff]
        _output_bed(src1_df, output_prefix+'.src1.bed', col='match_rate_x')
        _output_bed(src2_df, output_prefix+'.src2.bed', col='match_rate_y')


def _output_bed(df, output, col=None):
    """
    Description:
        Helper fuction for outputing candidate introgressed fragments in BED format.

    Arguments:
        df pandas.dataframe: Dataframe containing candidate introgressed fragments.
        output str: Name of the output file.
        col str: Name of the column containing source match rates in the dataframe.
    """
    df = df.sort_values(by=['sample', 'chrom', 'start', 'end'])
    if col is None:
        colnames = ['chrom', 'start', 'end', 'sample']
    else: colnames = ['chrom', 'start', 'end', 'sample', col]
    df.to_csv(output, columns=colnames, sep="\t", header=False, index=False)



# Apache License Version 2.0
# Copyright 2022 Xin Huang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import allel
import numpy as np
import os
from multiprocessing import Process, Queue
from sstar.utils import read_data, filter_data

#@profile
def cal_s_star(vcf, ref_ind_file, tgt_ind_file, anc_allele_file, output, win_len, win_step, thread, match_bonus, max_mismatch, mismatch_penalty):
    """
    Description:
        Calculate S* scores.
    
    Arguments:
        vcf str: Name of the VCF file containing genotype data.
        ref_ind_file str: Name of the file containing sample information from the reference population.
        tgt_ind_file str: Name of the file containing sample information from the target population.
        anc_allele_file str: Name of the file containing ancestral allele information.
        output str: Name of the output file.
        win_len int: Length of sliding windows.
        win_step int: Step size of sliding windows.
        thread int: Number of threads.
        match_bonus int: Bonus for matching genotypes of two different variants.
        max_mismatch int: Maximum genotype distance allowed.
        mismatch_penalty int: Penalty for mismatching genotypes of two different variants.
    """

    ref_data, ref_samples, tgt_data, tgt_samples, src_data, src_samples = read_data(vcf, ref_ind_file, tgt_ind_file, None, anc_allele_file)

    chr_names = tgt_data.keys()
    for c in chr_names:
        # Remove variants observed in the reference populations
        # Assume 1 is the alt allele
        variants_not_in_ref = np.sum(ref_data[c]['GT'].is_hom_ref(),axis=1) == len(ref_samples)
        tgt_data = filter_data(tgt_data, c, variants_not_in_ref)

    #header = 'chrom\tsample\tstart\tend\ttotal_SNP_number\tS*_SNP_number\tS*_score\tS*_SNPs'
    #o = open(output, 'w')
    #o.write(header+'\n')
    #for s in range(len(tgt_samples)):
    #    chr_names = tgt_data.keys()
    #    for c in chr_names:
    #        tgt_gt = tgt_data[c]['GT']
    #        tgt_pos = tgt_data[c]['POS']
    #        ind = tgt_gt[:,s]

    #        ref_gt = ref_data[c]['GT']
    #        ref_pos = ref_data[c]['POS']
    #        ref_sub_pos = ref_pos[~np.all(ref_gt.is_hom_ref(), axis=1)]
            # Assume the ref allele is 0 and the alt allele is 1
    #        tgt_sub_gt = tgt_gt[~ind.is_hom_ref()][:,s]
    #        tgt_sub_pos = tgt_pos[~ind.is_hom_ref()]
    #        res = _cal_score_ind(c, tgt_samples[s], ref_sub_pos, tgt_sub_pos, tgt_sub_gt, win_step, win_len)
    #        o.write('\n'.join(res))
    #o.close()

    if thread > 1: thread = min(os.cpu_count()-1, len(tgt_samples), thread)

    _cal_score(ref_data, tgt_data, tgt_samples, win_len, win_step, output, thread, match_bonus, max_mismatch, mismatch_penalty)

def _cal_score(ref_data, tgt_data, samples, win_len, win_step, output, thread, match_bonus, max_mismatch, mismatch_penalty):
    """
    Description:
        Helper function to manage worker functions to calculate S* scores with multiprocessing.

    Arguments:
        ref_data dict: Dictionary containing genotype data from the reference population.
        tgt_data dict: Dictionary containing genotype data from the target population.
        samples list: Sample information.
        win_len int: Length of sliding windows.
        win_step int: Step size of sliding windows.
        output str: Name of the output file.
        thread int: Number of threads.
        match_bonus int: Bonus for matching genotypes of two different variants.
        max_mismatch int: Maximum genotype distance allowed.
        mismatch_penalty int: Penalty for mismatching genotypes of two different variants.
    """
  
    try:
        from pytest_cov.embed import cleanup_on_sigterm
    except ImportError:
        pass
    else:
        cleanup_on_sigterm()
 
    res = []
    # Use multiprocessing to calculate S* scores in different samples
    in_queue, out_queue = Queue(), Queue()
    workers = [Process(target=_cal_score_worker, args=(in_queue, out_queue, ref_data, tgt_data, 
                       win_len, win_step, match_bonus, max_mismatch, mismatch_penalty)) for ii in range(thread)]
    for s in range(len(samples)):
        in_queue.put((s,samples[s]))

    try:
        for worker in workers:
            worker.start()

        for s in range(len(samples)):
            item = out_queue.get()
            if item != '': res.append(item)
    
        for worker in workers:
            worker.terminate()
    finally:
        for worker in workers:
            worker.join()

    header = 'chrom\tstart\tend\tsample\tS*_score\tregion_ind_SNP_number\tS*_SNP_number\tS*_SNPs'
    with open(output, 'w') as o:
        o.write(header+'\n')
        o.write('\n'.join(res))
        o.write('\n')

def _cal_score_worker(in_queue, out_queue, ref_data, tgt_data, win_len, win_step, match_bonus, max_mismatch, mismatch_penalty):
    """
    Description:
        Worker function to calculate S* scores with multiprocessing.

    Arguments:
        in_queue multiprocessing.Queue: multiprocessing.Queue instance to receive parameters from the manager.
        out_queue multiprocessing.Queue: multiprocessing.Queue instance to send results back to the manager.
        ref_data dict: Dictionary containing genotype data from the reference population.
        tgt_data dict: Dictionary containing genotype data from the target population.
        win_len int: Length of sliding windows.
        win_step int: Step size of sliding windows.
        match_bonus int: Bonus for matching genotypes of two different variants.
        max_mismatch int: Maximum genotype distance allowed.
        mismatch_penalty int: Penalty for mismatching genotypes of two different variants.
    """

    while True:
        s, sample_name = in_queue.get()
        chr_names = tgt_data.keys()
        for c in chr_names:
            tgt_gt = tgt_data[c]['GT']
            tgt_pos = tgt_data[c]['POS']
            ind = tgt_gt[:,s]

            ref_gt = ref_data[c]['GT']
            ref_pos = ref_data[c]['POS']
            ref_sub_pos = ref_pos[~np.all(ref_gt.is_hom_ref(), axis=1)]
            # Assume the ref allele is 0 and the alt allele is 1
            tgt_sub_gt = tgt_gt[~ind.is_hom_ref()][:,s]
            tgt_sub_pos = tgt_pos[~ind.is_hom_ref()]
            res = _cal_score_ind(c, sample_name, ref_sub_pos, tgt_sub_pos, tgt_sub_gt, win_step, win_len, match_bonus, max_mismatch, mismatch_penalty)
            out_queue.put('\n'.join(res))

#@profile
def _cal_score_ind(chr_name, sample_name, ref_pos, tgt_pos, tgt_gt, win_step, win_len, match_bonus, max_mismatch, mismatch_penalty):
    """
    Description:
        Helper function for calculating S* in a single individual.

    Arguments:
        chr_name str: Name of the chromosome.
        sample_name str: Name of the sample.
        ref_pos list: List of position for variants in reference populations.
        tgt_pos list: List of position for variants in target populations.
        tgt_gt allel.GenotypeArray: Genotype data from target populations.
        win_step int: Length of sliding windows.
        win_len int: Step size of sliding windows.
        match_bonus int: Bonus for matching genotypes of two different variants.
        max_mismatch int: Maximum genotype distance allowed.
        mismatch_penalty int: Penalty for mismatching genotypes of two different variants.

    Returns:
        res list: Results of calculated S*
    """
    res = []
    if len(tgt_gt) <= 2: 
        last_pos = 0 # An indivinual should have at least three SNPs for S* calculation
        win_start = 0
    else:
        win_start = (tgt_pos[0] + win_step) // win_step * win_step - win_len
        if win_start < 0: win_start = 0
        last_pos = tgt_pos[-1]

    while last_pos > win_start:
        win_end = win_start + win_len
        tgt_snps = np.where((tgt_pos>win_start) & (tgt_pos<=win_end))[0]
        ref_snps = np.where((ref_pos>win_start) & (ref_pos<=win_end))[0]
        total_snps = len(np.union1d(ref_pos[ref_snps], tgt_pos[tgt_snps]))

        # Only calculate S* for a region with more than two SNPs
        tgt_snp_num = len(tgt_snps)
        if tgt_snp_num > 2:
            tgt_snps_gt = tgt_gt[tgt_snps]
            tgt_snps_pos = tgt_pos[tgt_snps]
            # Store the maximum score ended at the j-th SNP
            max_scores = [0] * tgt_snp_num
            max_score_snps = [[]] * tgt_snp_num
            for j in range(tgt_snp_num):
                max_score = -np.inf
                snps = []
                for i in range(j):
                    geno_dist = abs(tgt_snps_gt[j][0] - tgt_snps_gt[i][0] + tgt_snps_gt[j][1] - tgt_snps_gt[i][1])
                    phy_dist = abs(tgt_snps_pos[j] - tgt_snps_pos[i])
                    if phy_dist < 10: score_ij = -np.inf
                    elif geno_dist == 0: score_ij = match_bonus + phy_dist
                    elif geno_dist <= max_mismatch: score_ij = mismatch_penalty
                    else: score_ij = -np.inf
                    score = max_scores[i] + score_ij
                    max_score_i = max(max_score, score, score_ij)
                    if max_score_i == max_score: continue
                    elif max_score_i == score: snps = max_score_snps[i] + [str(tgt_snps_pos[j])]
                    elif max_score_i == score_ij: snps = [str(tgt_snps_pos[i]), str(tgt_snps_pos[j])]
                    max_score = max_score_i
                max_scores[j] = max_score
                max_score_snps[j] = snps

            s_star_score = max(max_scores)
            last_snp = max_scores.index(s_star_score)
            haplotype = max_score_snps[last_snp]

            s_star_snp_num = len(haplotype)
            if len(haplotype) > 2:
                haplotype = ','.join(haplotype)
            else:
                s_star_snp_num = 'NA'
                s_star_score = 'NA'
                haplotype = 'NA'
        else:
            s_star_snp_num = 'NA'
            s_star_score = 'NA'
            haplotype = 'NA'
        
        line = f'{chr_name}\t{win_start}\t{win_end}\t{sample_name}\t{s_star_score}\t{total_snps}\t{s_star_snp_num}\t{haplotype}'
       
        win_start += win_step
        res.append(line)

    return res



# Apache License Version 2.0
# Copyright 2022 Xin Huang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import allel
import gzip
import os
import numpy as np
import pandas as pd
from multiprocessing import Process, Queue
from sstar.utils import read_data, py2round, read_mapped_region_file, cal_matchpct

#@profile
def cal_match_pct(vcf, ref_ind_file, tgt_ind_file, src_ind_file, anc_allele_file, output, thread, score_file, mapped_region_file):
    """
    Description:
        Calculate p-values for S* haplotypes in the target population with source genomes.

    Arguments:
        vcf str: Name of the VCF file containing genotypes.
        src_vcf str: Name of the VCF file containing genotypes from source populations.
        ref_ind_file str: Name of the file containing sample information from reference populations.
        tgt_ind_file str: Name of the file containing sample information from target populations.
        src_ind_file str: Name of the file containing sample information from source populations.
        anc_allele_file str: Name of the file containing ancestral allele information.
        output str: Name of the output file.
        thread int: Number of threads.
        score_file str: Name of the file containing S* scores calculated by `s-star score`.
        mapped_region_file str: Name of the BED file containing mapped regions.
    """

    ref_data, ref_samples, tgt_data, tgt_samples, src_data, src_samples = read_data(vcf, ref_ind_file, tgt_ind_file, src_ind_file, anc_allele_file)

    res = []
    chr_names = ref_data.keys()

    mapped_intervals = read_mapped_region_file(mapped_region_file)
    data, windows, samples = _read_score_file(score_file, chr_names, tgt_samples)
    sample_size = len(samples)

    header = 'chrom\tstart\tend\tsample\tmatch_rate\tsrc_sample'

    if thread > 1: thread = min(os.cpu_count()-1, sample_size, thread)    
    res = _cal_tgt_match_pct_manager(data, mapped_intervals, samples, tgt_samples, src_samples, tgt_data, src_data, sample_size, thread)

    with open(output, 'w') as o:
        o.write(header+"\n")
        o.write("\n".join(res)+"\n")

#@profile
def _read_score_file(score_file, chr_names, tgt_samples):
    """
    Description:
        Helper function for reading the file generated by `sstar score`.

    Arguments:
        score_file str: Name of the file containing S* scores generated by `sstar score`.
        chr_names list: List containing names of chromosomes for analysis.
        tgt_samples list: List containing names of samples from the target population for analysis.

    Returns:
        data dict: Dictionary containing S* for analysis.
        windows dict: Dictionary containing windows for analysis.
        header str: Header from the file generated by `sstar score`.
        samples list: List containing names of samples in the target population for analysis.
    """

    data = dict()
    windows = dict()
    for c in chr_names:
        windows[c] = []
    samples = []
    with open(score_file, 'r') as f:
        header = f.readline().rstrip()
        for line in f.readlines():
            line = line.rstrip()
            elements = line.split("\t")
            chr_name = elements[0]
            win_start = elements[1]
            win_end = elements[2]
            sample = elements[3]
            if sample not in tgt_samples: continue
            if elements[6] == 'NA': continue
            if sample not in data.keys(): 
                data[sample] = []
                samples.append(sample)
            data[sample].append(line)
            windows[c].append((int(win_start), int(win_end)))

    return data, windows, samples

def _cal_tgt_match_pct_manager(data, mapped_intervals, samples, tgt_samples, src_samples, tgt_data, src_data, sample_size, thread):
    """
    Description:
        Manager function to calculate match percents in target populations using multiprocessing.

    Arguments:
        data dict: Lines from the output file created by `sstar score`.
        mapped_intervals dict: Dictionary of tuples containing mapped regions across the genome.
        sample list: Sample information for individuals needed to be estimated match percents.
        tgt_samples list: Sample information from target populations.
        src_samples list: Sample information from source populations.
        tgt_data dict: Genotype data from target populations.
        src_data dict: Genotype data from source populations.
        sample_size int: Number of individuals analyzed.
        thread int: Number of threads.

    Returns:
        res list: Match percents for target populations.
    """

    try:
        from pytest_cov.embed import cleanup_on_sigterm
    except ImportError:
        pass
    else:
        cleanup_on_sigterm()

    res = []
    in_queue, out_queue = Queue(), Queue()
    workers = [Process(target=_cal_tgt_match_pct_worker, args=(in_queue, out_queue, mapped_intervals, tgt_data, src_data, src_samples, len(tgt_samples))) for ii in range(thread)]

    for t in samples:
        index = tgt_samples.index(t)
        in_queue.put((index, data[t]))
    
    try:
        for worker in workers:
            worker.start()
        for s in range(sample_size):
            item = out_queue.get()
            if item != '': res.append(item)
        for worker in workers:
            worker.terminate()
    finally:
        for worker in workers:
            worker.join()

    return res

def _cal_tgt_match_pct_worker(in_queue, out_queue, mapped_intervals, tgt_data, src_data, src_samples, sample_size):
    """
    Description:
        Worker function to calculate match percents in target populations.

    Arguments:
        in_queue multiprocessing.Queue: multiprocessing.Queue instance to receive parameters from the manager.
        out_queue multiprocessing.Queue: multiprocessing.Queue instance to send results back to the manager.
        mapped_intervals dict: Dictionary of tuples containing mapped regions across the genome.
        tgt_data dict: Genotype data from target populations.
        src_data dict: Genotype data from source populations.
        src_samples list: List containing sample information for source populations.
        sample_size int: Number of individuals analyzed.
    """

    while True:
        index, data = in_queue.get()
        res = _cal_match_pct_ind(data, index, mapped_intervals, tgt_data, src_data, src_samples, sample_size )
        out_queue.put("\n".join(res))

#@profile
def _cal_match_pct_ind(data, tgt_ind_index, mapped_intervals, tgt_data, src_data, src_samples, sample_size):
    """
    Description:
        Helper function for calculating p-values in individuals.

    Arguments:
        data dict: Dictionary containing S* for analysis.
        tgt_ind_index int: Index of the target individual for analysis.
        mapped_intervals dict: Dictionary of tuples containing mapped regions across the genome.
        tgt_data dict: Genotype data from target populations.
        src_data dict: Genotype data from source populations.
        src_samples list: List of samples from source populations.
        sample_size int: Number of individuals analyzed.

    Returns:
        res list: List containing estimated p-values and other statistics.
    """
    res = []
    for line in data:
        elements = line.split("\t")
        chr_name = elements[0]
        win_start, win_end = elements[1], elements[2]
        sample = elements[3]

        s_star_snps = elements[-1].split(",")
        s_start, s_end = s_star_snps[0], s_star_snps[-1]
        key1 = win_start+'-'+win_end
        key2 = s_start+'-'+s_end

        for src_ind_index in range(len(src_samples)):
            src_sample = src_samples[src_ind_index]
            hap1_res = cal_matchpct(chr_name, mapped_intervals, tgt_data, src_data, tgt_ind_index, src_ind_index, 0, int(win_start), int(win_end), sample_size)
            hap2_res = cal_matchpct(chr_name, mapped_intervals, tgt_data, src_data, tgt_ind_index, src_ind_index, 1, int(win_start), int(win_end), sample_size)
            
            hap1_match_pct = hap1_res[-1]
            hap2_match_pct = hap2_res[-1]
            hap_match_pct = 'NA'

            if (hap1_match_pct != 'NA') and (hap2_match_pct != 'NA'): hap_match_pct = (hap1_match_pct + hap2_match_pct) / 2

            line = f'{chr_name}\t{win_start}\t{win_end}\t{sample}\t{hap_match_pct}\t{src_sample}'
            res.append(line)
    return res



# Apache License Version 2.0
# Copyright 2022 Xin Huang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import allel
import math
import numpy as np

#@profile
def parse_ind_file(filename):
    """
    Description:
        Helper function to read sample information from files.

    Arguments:
        filename str: Name of the file containing sample information.

    Returns:
        samples list: Sample information.
    """
  
    f = open(filename, 'r') 
    samples = [l.rstrip() for l in f.readlines()] 
    f.close()

    if len(samples) == 0:
        raise Exception(f'No sample is found in {filename}! Please check your data.')

    return samples

#@profile
def read_geno_data(vcf, ind, anc_allele_file, filter_missing):
    """
    Description:
        Helper function to read genotype data from VCF files.

    Arguments:
        vcf str: Name of the VCF file containing genotype data.
        ind list: List containing names of samples.
        anc_allele_file str: Name of the BED file containing ancestral allele information.
        filter_missing bool: Indicating whether filtering missing data or not.

    Returns:
        data dict: Genotype data.
    """

    vcf = allel.read_vcf(vcf, alt_number=1, samples=ind)
    gt = vcf['calldata/GT']
    chr_names = np.unique(vcf['variants/CHROM'])
    samples = vcf['samples']
    pos = vcf['variants/POS']
    ref = vcf['variants/REF']
    alt = vcf['variants/ALT']

    if anc_allele_file != None: anc_allele = read_anc_allele(anc_allele_file)
    data = dict()
    for c in chr_names:
        if c not in data.keys():
            data[c] = dict()
            data[c]['POS'] = pos
            data[c]['REF'] = ref
            data[c]['ALT'] = alt
            data[c]['GT'] = gt
        index = np.where(vcf['variants/CHROM'] == c)
        data = filter_data(data, c, index)
        # Remove missing data
        if filter_missing:
            index = data[c]['GT'].count_missing(axis=1) == len(samples)
            data = filter_data(data, c, ~index)
        if anc_allele_file != None: data = check_anc_allele(data, anc_allele, c)

    return data

#@profile
def filter_data(data, c, index):
    """
    Description:
        Helper function to filter genotype data.

    Arguments:
        data dict: Genotype data for filtering.
            c str: Names of chromosomes.
        index numpy.ndarray: A boolean array determines variants to be removed.

    Returns:
        data dict: Genotype data after filtering.
    """

    data[c]['POS'] = data[c]['POS'][index]
    data[c]['REF'] = data[c]['REF'][index]
    data[c]['ALT'] = data[c]['ALT'][index]
    data[c]['GT'] = allel.GenotypeArray(data[c]['GT'][index])

    return data

#@profile
def read_data(vcf, ref_ind_file, tgt_ind_file, src_ind_file, anc_allele_file):
    """
    Description:
        Helper function for reading data from reference and target populations.

    Arguments:
        vcf str: Name of the VCF file containing genotype data from reference, target, and source populations.
        ref_ind_file str: Name of the file containing sample information from reference populations.
        tgt_ind_file str: Name of the file containing sample information from target populations.
        src_ind_file str: Name of the file containing sample information from source populations.
        anc_allele_file str: Name of the file containing ancestral allele information.

    Returns:
        ref_data dict: Genotype data from reference populations.
        ref_samples list: Sample information from reference populations.
        tgt_data dict: Genotype data from target populations.
        tgt_samples list: Sample information from target populations.
        src_data list: Genotype data from source populations.
        src_samples list: Sample information from source populations.
    """
   
    ref_data = ref_samples = tgt_data = tgt_samples = src_data = src_samples = None
    if ref_ind_file != None: 
        ref_samples = parse_ind_file(ref_ind_file)
        ref_data = read_geno_data(vcf, ref_samples, anc_allele_file, True)
        
    if tgt_ind_file != None: 
        tgt_samples = parse_ind_file(tgt_ind_file)
        tgt_data = read_geno_data(vcf, tgt_samples, anc_allele_file, True)

    if src_ind_file != None:
        src_samples = parse_ind_file(src_ind_file)
        src_data = read_geno_data(vcf, src_samples, anc_allele_file, False)

    if (ref_ind_file != None) and (tgt_ind_file != None):
        chr_names = tgt_data.keys()
        for c in chr_names:
            # Remove variants fixed in both the reference and target individuals
            ref_fixed_variants = np.sum(ref_data[c]['GT'].is_hom_alt(),axis=1) == len(ref_samples)
            tgt_fixed_variants = np.sum(tgt_data[c]['GT'].is_hom_alt(),axis=1) == len(tgt_samples)
            fixed_index = np.logical_and(ref_fixed_variants, tgt_fixed_variants)
            index = np.logical_not(fixed_index)
            fixed_pos =ref_data[c]['POS'][fixed_index]
            ref_data = filter_data(ref_data, c, index)
            tgt_data = filter_data(tgt_data, c, index)
            if src_ind_file != None:
                index = np.logical_not(np.in1d(src_data[c]['POS'], fixed_pos))
                src_data = filter_data(src_data, c, index)

    return ref_data, ref_samples, tgt_data, tgt_samples, src_data, src_samples

#@profile
def get_ref_alt_allele(ref, alt, pos):
    """
    Description:
        Helper function to index REF and ALT alleles with genomic positions.

    Arguments:
        ref list: REF alleles.
        alt list: ALT alleles.
        pos list: Genomic positions.

    Returns:
        ref_allele dict: REF alleles.
        alt_allele dict: ALT alleles.
    """
    
    ref_allele = dict()
    alt_allele = dict()

    for i in range(len(pos)):
        r = ref[i]
        a = alt[i]
        p = pos[i]
        ref_allele[p] = r
        alt_allele[p] = a
   
    return ref_allele, alt_allele

#@profile
def read_anc_allele(anc_allele_file):
    """
    Description:
        Helper function to read ancestral allele information from files.

    Arguments:
        anc_allele_file str: Name of the BED file containing ancestral allele information.

    Returns:
        anc_allele dict: Ancestral allele information.
    """

    anc_allele = dict()
    with open(anc_allele_file, 'r') as f:
        for line in f.readlines():
            e = line.rstrip().split()
            if e[0] not in anc_allele: anc_allele[e[0]] = dict()
            anc_allele[e[0]][int(e[2])] = e[3]

    if not anc_allele: raise Exception(f'No ancestral allele is found! Please check your data.')
    
    return anc_allele

#@profile
def check_anc_allele(data, anc_allele, c):
    """
    Description:
        Helper function to check whether the REF or ALT allele is the ancestral allele.
        If the ALT allele is the ancestral allele, then the genotypes in this position will be flipped.
        If neither the REF nor ALT allele is the ancestral allele, then this position will be removed.
        If a position has no the ancestral allele information, the this position will be removed.

    Arguments:
        data dict: Genotype data for checking ancestral allele information.
        anc_allele dict: Ancestral allele information for checking.
        c str: Name of the chromosome.

    Returns:
        data dict: Genotype data after checking.
    """

    ref_allele, alt_allele = get_ref_alt_allele(data[c]['REF'], data[c]['ALT'], data[c]['POS'])
    # Remove variants not in the ancestral allele file
    intersect_snps = np.intersect1d(list(ref_allele.keys()), list(anc_allele[c].keys()))
    # Remove variants that neither the ref allele nor the alt allele is the ancestral allele
    removed_snps = []
    # Flip variants that the alt allele is the ancestral allele
    flipped_snps = []

    for v in intersect_snps:
        if (anc_allele[c][v] != ref_allele[v]) and (anc_allele[c][v] != alt_allele[v]): removed_snps.append(v)
        elif (anc_allele[c][v] == alt_allele[v]): flipped_snps.append(v)

    intersect_snps = np.in1d(data[c]['POS'], intersect_snps)
    data = filter_data(data, c, intersect_snps)

    if len(removed_snps) != 0:
        remained_snps = np.logical_not(np.in1d(data[c]['POS'], removed_snps))
        data = filter_data(data, c, remained_snps)

    is_flipped_snps = np.in1d(data[c]['POS'], flipped_snps)
    # Assume no missing data
    for i in range(len(data[c]['POS'])):
        if is_flipped_snps[i]:
            data[c]['GT'][i] = allel.GenotypeVector(abs(data[c]['GT'][i]-1))

    return data

#@profile
def read_mapped_region_file(mapped_region):
    """
    Description:
        Helper function for reading mapped regions from a BED file.

    Arguments:
        mapped_region str: BED file containing mapped regions.

    Returns:
        mapped_intervals dict: Dictionary of tuples containing mapped regions.
    """

    if mapped_region != None:
        mapped_intervals = dict()
        with open(mapped_region, 'r') as m:
            for line in m.readlines():
                elements = line.rstrip().split("\t")
                if elements[0] not in mapped_intervals.keys(): mapped_intervals[elements[0]] = []
                mapped_intervals[elements[0]].append((int(elements[1]), int(elements[2])))
    else: mapped_intervals = None

    return mapped_intervals

#@profile
def cal_matchpct(chr_name, mapped_intervals, data, src_data, tgt_ind_index, src_ind_index, hap_index, win_start, win_end, sample_size):
    """
    Description:
        Helper function to calculate match percents in a given individual.

    Arguments:
        chr_name str: Name of the chromosomes.
        mapped_intervals: Dictionary of tuples containing mapped regions across the genome.
        data dict: Genotype data from individuals to be calculated match percents.
        src_data dict: Genotype data from source populations.
        tgt_ind_index int: Index of the target individual for calculating match percent.
        src_ind_index int: Index of the source individual for calculating match percent.
        hap_index int: Index of the haplotype for calculating match percent.
        win_start int: Start position of the window for calculating match percent.
        win_end int: End position of the window for calculating match percent.
        sample_size int: Number of individuals analyzed.

    Returns:
        res list: List containing statistics for the given haplotype.
    """

    if (win_start == 'NA') and (win_end == 'NA'): 
        return ['NA', 'NA', 'NA', 'NA', 'NA', 'NA', 'NA']

    win_start = int(win_start)
    win_end = int(win_end)

    hap_match_pct = 'NA'
    gt = data[chr_name]['GT']
    pos = data[chr_name]['POS']
    src_gt = src_data[chr_name]['GT']
    src_pos = src_data[chr_name]['POS']
    res = []
    
    sub_snps = np.where((pos>=win_start) & (pos<=win_end))[0]
    sub_gt = gt[sub_snps][:,tgt_ind_index]
    sub_pos = pos[sub_snps]

    sub_src_snps = np.where((src_pos>=win_start) & (src_pos<=win_end))[0]
    sub_src_gt = src_gt[sub_src_snps][:,src_ind_index]
    sub_src_pos = src_pos[sub_src_snps]
    missing_index = sub_src_gt.is_missing()
    sub_src_gt = sub_src_gt[~missing_index]
    sub_src_pos = sub_src_pos[~missing_index]
    src_hom_variants = sub_src_pos[sub_src_gt.is_hom_alt()]
    src_het_variants = sub_src_pos[sub_src_gt.is_het()]
    src_variants = np.concatenate((src_hom_variants, src_het_variants))

    hap = sub_gt[:,hap_index]
    hap_len = win_end - win_start
    hap_mapped_len = _cal_mapped_len(mapped_intervals, chr_name, win_start, win_end)
    hap_variants_num, hap_site_num, hap_match_src_allele_num, hap_sfs, hap_match_pct = _cal_hap_stats(gt[sub_snps], hap, sub_pos, src_variants, src_hom_variants, src_het_variants, sample_size)

    res.append(hap_variants_num)
    res.append(hap_len)
    res.append(hap_mapped_len)
    res.append(hap_match_src_allele_num)
    res.append(hap_site_num)
    res.append(hap_sfs)
    res.append(hap_match_pct)

    return res

#@profile
def _cal_mapped_len(mapped_intervals, chr_name, win_start, win_end):
    """
    Description:
        Helper function for calculating length of mapped region in a given window.

    Arguments:
        mapped_intervals dict: Dictionary of tuples containing mapped regions across the genome.
        chr_name str: Name of the chromosome for calculating the length of the mapped region.
        win_start int: Start position of a window.
        win_end int: End position of a window.

    Returns:
        mapped_len int: Length of the mapped region.
    """

    if (mapped_intervals == None) or (chr_name not in mapped_intervals.keys()):
        mapped_len = win_end - win_start
    else:
        overlaps = [(idx[0], idx[1]) for idx in mapped_intervals[chr_name] if (win_start>=idx[0] and win_start<=idx[1]) or (win_end>=idx[0] and win_end<=idx[1]) or (win_start<=idx[0] and win_end>=idx[1])]
        mapped_len = 0
        for idx in overlaps:
            if (idx[0]<=win_start) and (idx[1]>=win_end): mapped_len += win_end - win_start
            elif (idx[0]>=win_start) and (idx[1]<=win_end): mapped_len += idx[1] - idx[0]
            elif (idx[0]<=win_start) and (idx[1]<=win_end): mapped_len += idx[1] - win_start
            else: mapped_len += win_end - idx[0]

    mapped_len = mapped_len // 1000 * 1000
        
    return mapped_len
     
#@profile
def _cal_hap_stats(gt, hap, pos, src_variants, src_hom_variants, src_het_variants, sample_size):
    """
    Description:
        Helper function for calculating statistics for a haplotype.

    Arguments:
        gt allel.GenotypeArray: Genotype data for all the haplotypes within the same window of the haplotype to be analyzed.
        hap allel.GenotypeVector: Genotype data for the haplotype to be analyzed.
        pos list: List containing positions of variants on the haplotype.
        src_variants list: List containing positions of variants on the individual from the source population.
        src_hom_variants list: List containing positions of homozygous variants on the individual from the source population.
        src_het_variants list: List containing positions of heterozygous variants on the individual from the source population.
        sample_size int: Number of individuals analyzed.

    Returns:
        hap_variants_num int: Number of SNPs with derived alleles on the haplotype.
        hap_site_num int: Number of SNPs with derived alleles either on the haplotype or the source genomes.
        hap_match_src_allele_num int: Number of SNPs with derived alleles both on the haplotype and the source genomes.
        hap_sfs int: Average number of derived variants per site per haplotype.
        hap_match_pct float: Match percent of the haplotype.
        sample_size int: Number of individuals analyzed.
    """

    if hap is None: return 'NA', 'NA', 'NA', 'NA', 'NA'
    else:
        hap_variants = pos[np.equal(hap, 1)]
        hap_variants_num = len(hap_variants)
        # Assume the alternative allele is the derived allele
        hap_shared_src_hom_site_num = len(np.intersect1d(hap_variants, src_hom_variants))
        hap_shared_src_het_site_num = len(np.intersect1d(hap_variants, src_het_variants))
        hap_site_num = len(np.union1d(hap_variants, src_variants))
        hap_match_src_allele_num = hap_shared_src_hom_site_num + 0.5*hap_shared_src_het_site_num
        hap_shared_src_site_num = hap_shared_src_hom_site_num + hap_shared_src_het_site_num
        if hap_site_num != 0: hap_match_pct = round(hap_match_src_allele_num/hap_site_num, 6)
        else: hap_match_pct = 'NA'

        hap_sfs = np.sum(np.sum(gt[hap == 1], axis=2), axis=1)
        if hap_sfs.size != 0:
            hap_sfs_mean = np.mean(hap_sfs)
            # See https://stackoverflow.com/questions/10825926/python-3-x-rounding-behavior
            #if not np.isnan(sfs_mean): sfs_mean = int(round(sfs_mean))
            #if not np.isnan(hap_sfs_mean): hap_sfs = int(int(py2round(hap_sfs_mean))/10*108)
            #if not np.isnan(hap_sfs_mean): hap_sfs = int(py2round(hap_sfs_mean))/(2*sample_size)
            if not np.isnan(hap_sfs_mean): hap_sfs = round(hap_sfs_mean/(2*sample_size), 6)
        else:
            hap_sfs = np.nan

    return hap_variants_num, hap_site_num, hap_match_src_allele_num, hap_sfs, hap_match_pct

def py2round(x, d=0):
    p = 10 ** d
    if x > 0: return float(math.floor((x * p) + 0.5))/p
    elif x < 0: float(math.ceil((x * p) - 0.5))/p
    else: return 0.0
